/*
 * Copyright (c) 2014, 2018, Oracle and/or its affiliates. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License, version 2.0,
 * as published by the Free Software Foundation.
 *
 * This program is designed to work with certain software (including
 * but not limited to OpenSSL) that is licensed under separate terms, as
 * designated in a particular file or component or in included license
 * documentation.  The authors of MySQL hereby grant you an additional
 * permission to link the program and your derivative works with the
 * separately licensed software that they have either included with
 * the program or referenced in the documentation.
 * This program is distributed in the hope that it will be useful,  but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License, version 2.0, for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA 
 */

#pragma once

#include "mforms/form.h"

#include "grts/structs.db.mysql.h"

namespace mforms {
  class Button;
  class TextEntry;
}

class ExecuteRoutineWizard : public mforms::Form {
private:
  db_mysql_RoutineRef _routine;
  db_mysql_CatalogRef _catalog;
  std::vector<mforms::TextEntry *> _edits;
  std::string _sql_mode;

  mforms::Button *_cancel_button;
  mforms::Button *_execxute_button;

  bool needs_quoting(const std::string &type);

public:
  ExecuteRoutineWizard(db_mysql_RoutineRef routine, const std::string &sql_mode);

  std::string run();
};
