#ifndef _SQL_PARSER_SYMBOLS_H_
#define _SQL_PARSER_SYMBOLS_H_

namespace sql
{

enum symbol
{
_,
_123,
_124,
_125,
_126,
_33,
_37,
_38,
_40,
_41,
_42,
_43,
_44,
_45,
_46,
_47,
_58,
_59,
_64,
_94,
_ACCESSIBLE_SYM,
_ACTION,
_ADD,
_ADDDATE_SYM,
_AFTER_SYM,
_AGAINST,
_AGGREGATE_SYM,
_ALGORITHM_SYM,
_ALL,
_ALTER,
_ANALYSE_SYM,
_ANALYZE_SYM,
_AND_AND_SYM,
_AND_SYM,
_ANY_SYM,
_AS,
_ASC,
_ASCII_SYM,
_AT_SYM,
_AUTHORS_SYM,
_AUTOEXTEND_SIZE_SYM,
_AUTO_INC,
_AVG_ROW_LENGTH,
_AVG_SYM,
_BACKUP_SYM,
_BEFORE_SYM,
_BEGIN_SYM,
_BETWEEN_SYM,
_BIGINT,
_BINARY,
_BINLOG_SYM,
_BIN_NUM,
_BIT_AND,
_BIT_OR,
_BIT_SYM,
_BIT_XOR,
_BLOB_SYM,
_BLOCK_SYM,
_BOOLEAN_SYM,
_BOOL_SYM,
_BOTH,
_BTREE_SYM,
_BY,
_BYTE_SYM,
_CACHE_SYM,
_CALL_SYM,
_CASCADE,
_CASCADED,
_CASE_SYM,
_CAST_SYM,
_CATALOG_NAME_SYM,
_CHAIN_SYM,
_CHANGE,
_CHANGED,
_CHARSET,
_CHAR_SYM,
_CHECKSUM_SYM,
_CHECK_SYM,
_CIPHER_SYM,
_CLASS_ORIGIN_SYM,
_CLIENT_SYM,
_CLOSE_SYM,
_COALESCE,
_CODE_SYM,
_COLLATE_SYM,
_COLLATION_SYM,
_COLUMNS,
_COLUMN_FORMAT_SYM,
_COLUMN_NAME_SYM,
_COLUMN_SYM,
_COMMENT_SYM,
_COMMITTED_SYM,
_COMMIT_SYM,
_COMPACT_SYM,
_COMPLETION_SYM,
_COMPRESSED_SYM,
_CONCURRENT,
_CONDITION_SYM,
_CONNECTION_SYM,
_CONSISTENT_SYM,
_CONSTRAINT,
_CONSTRAINT_CATALOG_SYM,
_CONSTRAINT_NAME_SYM,
_CONSTRAINT_SCHEMA_SYM,
_CONTAINS_SYM,
_CONTEXT_SYM,
_CONTINUE_SYM,
_CONTRIBUTORS_SYM,
_CONVERT_SYM,
_COUNT_SYM,
_CPU_SYM,
_CREATE,
_CROSS,
_CUBE_SYM,
_CURDATE,
_CURRENT_SYM,
_CURRENT_USER,
_CURSOR_NAME_SYM,
_CURSOR_SYM,
_CURTIME,
_DATABASE,
_DATABASES,
_DATAFILE_SYM,
_DATA_SYM,
_DATETIME,
_DATE_ADD_INTERVAL,
_DATE_SUB_INTERVAL,
_DATE_SYM,
_DAY_HOUR_SYM,
_DAY_MICROSECOND_SYM,
_DAY_MINUTE_SYM,
_DAY_SECOND_SYM,
_DAY_SYM,
_DEALLOCATE_SYM,
_DECIMAL_NUM,
_DECIMAL_SYM,
_DECLARE_SYM,
_DEFAULT,
_DEFAULT_AUTH_SYM,
_DEFINER_SYM,
_DELAYED_SYM,
_DELAY_KEY_WRITE_SYM,
_DELETE_SYM,
_DESC,
_DESCRIBE,
_DES_KEY_FILE,
_DETERMINISTIC_SYM,
_DIAGNOSTICS_SYM,
_DIRECTORY_SYM,
_DISABLE_SYM,
_DISCARD,
_DISK_SYM,
_DISTINCT,
_DIV_SYM,
_DOUBLE_SYM,
_DO_SYM,
_DROP,
_DUAL_SYM,
_DUMPFILE,
_DUPLICATE_SYM,
_DYNAMIC_SYM,
_EACH_SYM,
_EDIT_SYM,
_ELSE,
_ELSEIF_SYM,
_ENABLE_SYM,
_ENCLOSED,
_END,
_ENDS_SYM,
_END_OF_INPUT,
_ENGINES_SYM,
_ENGINE_SYM,
_ENUM,
_EQ,
_EQUAL_SYM,
_ERRORS,
_ERROR_SYM,
_ESCAPED,
_ESCAPE_SYM,
_EVENTS_SYM,
_EVENT_SYM,
_EVERY_SYM,
_EXCHANGE_SYM,
_EXECUTE_SYM,
_EXISTS,
_EXIT_SYM,
_EXPANSION_SYM,
_EXPIRE_SYM,
_EXPORT_SYM,
_EXTENDED_SYM,
_EXTENT_SIZE_SYM,
_EXTRACT_SYM,
_FALSE_SYM,
_FAST_SYM,
_FAULTS_SYM,
_FETCH_SYM,
_FILE_SYM,
_FIRST_SYM,
_FIXED_SYM,
_FLOAT_NUM,
_FLOAT_SYM,
_FLUSH_SYM,
_FORCE_SYM,
_FOREIGN,
_FORMAT_SYM,
_FOR_SYM,
_FOUND_SYM,
_FROM,
_FULL,
_FULLTEXT_SYM,
_FUNCTION_SYM,
_GE,
_GENERAL,
_GEOMETRYCOLLECTION,
_GEOMETRY_SYM,
_GET_FORMAT,
_GET_SYM,
_GLOBAL_SYM,
_GRANT,
_GRANTS,
_GROUP_CONCAT_SYM,
_GROUP_SYM,
_GT_SYM,
_HANDLER_SYM,
_HASH_SYM,
_HAVING,
_HELP_SYM,
_HEX_NUM,
_HIGH_PRIORITY,
_HOSTS_SYM,
_HOST_SYM,
_HOUR_MICROSECOND_SYM,
_HOUR_MINUTE_SYM,
_HOUR_SECOND_SYM,
_HOUR_SYM,
_IDENT,
_IDENTIFIED_SYM,
_IDENT_QUOTED,
_IDENT_sys,
_IF,
_IGNORE_SERVER_IDS_SYM,
_IGNORE_SYM,
_IMPORT,
_INDEXES,
_INDEX_SYM,
_INFILE,
_INITIAL_SIZE_SYM,
_INNER_SYM,
_INOUT_SYM,
_INSERT,
_INSERT_METHOD,
_INSTALL_SYM,
_INTERVAL_SYM,
_INTO,
_INT_SYM,
_INVOKER_SYM,
_IN_SYM,
_IO_SYM,
_IPC_SYM,
_IS,
_ISOLATION,
_ISSUER_SYM,
_ITERATE_SYM,
_JOIN_SYM,
_KEYS,
_KEY_BLOCK_SIZE,
_KEY_SYM,
_KILL_SYM,
_LANGUAGE_SYM,
_LAST_SYM,
_LE,
_LEADING,
_LEAVES,
_LEAVE_SYM,
_LEFT,
_LESS_SYM,
_LEVEL_SYM,
_LEX_HOSTNAME,
_LIKE,
_LIMIT,
_LINEAR_SYM,
_LINES,
_LINESTRING,
_LIST_SYM,
_LOAD,
_LOCAL_SYM,
_LOCKS_SYM,
_LOCK_SYM,
_LOGFILE_SYM,
_LOGS_SYM,
_LONGBLOB,
_LONGTEXT,
_LONG_NUM,
_LONG_SYM,
_LOOP_SYM,
_LOW_PRIORITY,
_LT,
_MASTER_AUTO_POSITION_SYM,
_MASTER_BIND_SYM,
_MASTER_CONNECT_RETRY_SYM,
_MASTER_DELAY_SYM,
_MASTER_HEARTBEAT_PERIOD_SYM,
_MASTER_HOST_SYM,
_MASTER_LOG_FILE_SYM,
_MASTER_LOG_POS_SYM,
_MASTER_PASSWORD_SYM,
_MASTER_PORT_SYM,
_MASTER_RETRY_COUNT_SYM,
_MASTER_SERVER_ID_SYM,
_MASTER_SSL_CAPATH_SYM,
_MASTER_SSL_CA_SYM,
_MASTER_SSL_CERT_SYM,
_MASTER_SSL_CIPHER_SYM,
_MASTER_SSL_CRLPATH_SYM,
_MASTER_SSL_CRL_SYM,
_MASTER_SSL_KEY_SYM,
_MASTER_SSL_SYM,
_MASTER_SSL_VERIFY_SERVER_CERT_SYM,
_MASTER_SYM,
_MASTER_USER_SYM,
_MATCH,
_MAX_CONNECTIONS_PER_HOUR,
_MAX_QUERIES_PER_HOUR,
_MAX_ROWS,
_MAX_SIZE_SYM,
_MAX_SYM,
_MAX_UPDATES_PER_HOUR,
_MAX_USER_CONNECTIONS_SYM,
_MAX_VALUE_SYM,
_MEDIUMBLOB,
_MEDIUMINT,
_MEDIUMTEXT,
_MEDIUM_SYM,
_MEMORY_SYM,
_MERGE_SYM,
_MESSAGE_TEXT_SYM,
_MICROSECOND_SYM,
_MIGRATE_SYM,
_MINUTE_MICROSECOND_SYM,
_MINUTE_SECOND_SYM,
_MINUTE_SYM,
_MIN_ROWS,
_MIN_SYM,
_MODE_SYM,
_MODIFIES_SYM,
_MODIFY_SYM,
_MOD_SYM,
_MONTH_SYM,
_MULTILINESTRING,
_MULTIPOINT,
_MULTIPOLYGON,
_MUTEX_SYM,
_MYSQL_ERRNO_SYM,
_NAMES_SYM,
_NAME_SYM,
_NATIONAL_SYM,
_NATURAL,
_NCHAR_STRING,
_NCHAR_SYM,
_NDBCLUSTER_SYM,
_NE,
_NEG,
_NEW_SYM,
_NEXT_SYM,
_NODEGROUP_SYM,
_NONE_SYM,
_NOT2_SYM,
_NOT_SYM,
_NOW_SYM,
_NO_SYM,
_NO_WAIT_SYM,
_NO_WRITE_TO_BINLOG,
_NULL_SYM,
_NUM,
_NUMBER_SYM,
_NUMERIC_SYM,
_NUM_literal,
_NVARCHAR_SYM,
_OFFSET_SYM,
_OLD_PASSWORD,
_ON,
_ONE_SYM,
_ONLY_SYM,
_OPEN_SYM,
_OPTIMIZE,
_OPTION,
_OPTIONALLY,
_OPTIONS_SYM,
_OR2_SYM,
_ORDER_SYM,
_OR_OR_SYM,
_OR_SYM,
_OUTER,
_OUTFILE,
_OUT_SYM,
_OWNER_SYM,
_PACK_KEYS_SYM,
_PAGE_SYM,
_PARAM_MARKER,
_PARSER_SYM,
_PARTIAL,
_PARTITIONING_SYM,
_PARTITIONS_SYM,
_PARTITION_SYM,
_PASSWORD,
_PHASE_SYM,
_PLUGINS_SYM,
_PLUGIN_DIR_SYM,
_PLUGIN_SYM,
_POINT_SYM,
_POLYGON,
_PORT_SYM,
_POSITION_SYM,
_PRECISION,
_PREPARE_SYM,
_PRESERVE_SYM,
_PREV_SYM,
_PRIMARY_SYM,
_PRIVILEGES,
_PROCEDURE_SYM,
_PROCESS,
_PROCESSLIST_SYM,
_PROFILES_SYM,
_PROFILE_SYM,
_PROXY_SYM,
_PURGE,
_QUARTER_SYM,
_QUERY_SYM,
_QUICK,
_RANGE_SYM,
_READS_SYM,
_READ_ONLY_SYM,
_READ_SYM,
_READ_WRITE_SYM,
_REAL,
_REBUILD_SYM,
_RECOVER_SYM,
_REDOFILE_SYM,
_REDO_BUFFER_SIZE_SYM,
_REDUNDANT_SYM,
_REFERENCES,
_REGEXP,
_RELAY,
_RELAYLOG_SYM,
_RELAY_LOG_FILE_SYM,
_RELAY_LOG_POS_SYM,
_RELAY_THREAD,
_RELEASE_SYM,
_RELOAD,
_REMOVE_SYM,
_RENAME,
_REORGANIZE_SYM,
_REPAIR,
_REPEATABLE_SYM,
_REPEAT_SYM,
_REPLACE,
_REPLICATION,
_REQUIRE_SYM,
_RESET_SYM,
_RESIGNAL_SYM,
_RESOURCES,
_RESTORE_SYM,
_RESTRICT,
_RESUME_SYM,
_RETURNED_SQLSTATE_SYM,
_RETURNS_SYM,
_RETURN_SYM,
_REVERSE_SYM,
_REVOKE,
_RIGHT,
_ROLLBACK_SYM,
_ROLLUP_SYM,
_ROUTINE_SYM,
_ROWS_SYM,
_ROW_COUNT_SYM,
_ROW_FORMAT_SYM,
_ROW_SYM,
_RTREE_SYM,
_SAVEPOINT_SYM,
_SCHEDULE_SYM,
_SCHEMA_NAME_SYM,
_SECOND_MICROSECOND_SYM,
_SECOND_SYM,
_SECURITY_SYM,
_SELECT_SYM,
_SEPARATOR_SYM,
_SERIALIZABLE_SYM,
_SERIAL_SYM,
_SERVER_SYM,
_SESSION_SYM,
_SET,
_SET_VAR,
_SHARE_SYM,
_SHIFT_LEFT,
_SHIFT_RIGHT,
_SHOW,
_SHUTDOWN,
_SIGNAL_SYM,
_SIGNED_SYM,
_SIMPLE_SYM,
_SLAVE,
_SLOW,
_SMALLINT,
_SNAPSHOT_SYM,
_SOCKET_SYM,
_SONAME_SYM,
_SOUNDS_SYM,
_SOURCE_SYM,
_SPATIAL_SYM,
_SQLEXCEPTION_SYM,
_SQLSTATE_SYM,
_SQLWARNING_SYM,
_SQL_AFTER_GTIDS,
_SQL_AFTER_MTS_GAPS,
_SQL_BEFORE_GTIDS,
_SQL_BIG_RESULT,
_SQL_BUFFER_RESULT,
_SQL_CACHE_SYM,
_SQL_CALC_FOUND_ROWS,
_SQL_NO_CACHE_SYM,
_SQL_SMALL_RESULT,
_SQL_SYM,
_SQL_THREAD,
_SSL_SYM,
_STARTING,
_STARTS_SYM,
_START_SYM,
_STATS_AUTO_RECALC_SYM,
_STATS_PERSISTENT_SYM,
_STATS_SAMPLE_PAGES_SYM,
_STATUS_SYM,
_STDDEV_SAMP_SYM,
_STD_SYM,
_STOP_SYM,
_STORAGE_SYM,
_STRAIGHT_JOIN,
_STRING_SYM,
_SUBCLASS_ORIGIN_SYM,
_SUBDATE_SYM,
_SUBJECT_SYM,
_SUBPARTITIONS_SYM,
_SUBPARTITION_SYM,
_SUBSTRING,
_SUM_SYM,
_SUPER_SYM,
_SUSPEND_SYM,
_SWAPS_SYM,
_SWITCHES_SYM,
_SYSDATE,
_TABLES,
_TABLESPACE,
_TABLE_CHECKSUM_SYM,
_TABLE_NAME_SYM,
_TABLE_REF_PRIORITY,
_TABLE_SYM,
_TEMPORARY,
_TEMPTABLE_SYM,
_TERMINATED,
_TEXT_STRING,
_TEXT_STRING_filesystem,
_TEXT_STRING_literal,
_TEXT_STRING_sys,
_TEXT_STRING_sys_nonewline,
_TEXT_SYM,
_THAN_SYM,
_THEN_SYM,
_TIMESTAMP,
_TIMESTAMP_ADD,
_TIMESTAMP_DIFF,
_TIME_SYM,
_TINYBLOB,
_TINYINT,
_TINYTEXT,
_TO_SYM,
_TRAILING,
_TRANSACTION_SYM,
_TRIGGERS_SYM,
_TRIGGER_SYM,
_TRIM,
_TRUE_SYM,
_TRUNCATE_SYM,
_TYPES_SYM,
_TYPE_SYM,
_UDF_RETURNS_SYM,
_ULONGLONG_NUM,
_UNCOMMITTED_SYM,
_UNDEFINED_SYM,
_UNDERSCORE_CHARSET,
_UNDOFILE_SYM,
_UNDO_BUFFER_SIZE_SYM,
_UNICODE_SYM,
_UNINSTALL_SYM,
_UNION_SYM,
_UNIQUE_SYM,
_UNKNOWN_SYM,
_UNLOCK_SYM,
_UNSIGNED,
_UNTIL_SYM,
_UPDATE_SYM,
_UPGRADE_SYM,
_USAGE,
_USER,
_USE_FRM,
_USE_SYM,
_USING,
_UTC_DATE_SYM,
_UTC_TIMESTAMP_SYM,
_UTC_TIME_SYM,
_VALUES,
_VALUE_SYM,
_VARBINARY,
_VARCHAR,
_VARIABLES,
_VARIANCE_SYM,
_VARYING,
_VAR_SAMP_SYM,
_VIEW_SYM,
_WAIT_SYM,
_WARNINGS,
_WEEK_SYM,
_WEIGHT_STRING_SYM,
_WHEN_SYM,
_WHERE,
_WHILE_SYM,
_WITH,
_WITH_CUBE_SYM,
_WITH_ROLLUP_SYM,
_WORK_SYM,
_WRAPPER_SYM,
_WRITE_SYM,
_X509_SYM,
_XA_SYM,
_XML_SYM,
_XOR,
_YEAR_MONTH_SYM,
_YEAR_SYM,
_ZEROFILL,
_add_column,
_add_log_file,
_add_part_extra,
_add_partition_rule,
_adm_partition,
_all_key_opt,
_all_or_alt_part_name_list,
_all_or_any,
_alt_part_name_item,
_alt_part_name_list,
_alter,
_alter_algorithm_option,
_alter_commands,
_alter_list,
_alter_list_item,
_alter_lock_option,
_alter_logfile_group_info,
_alter_logfile_group_option,
_alter_logfile_group_option_list,
_alter_logfile_group_options,
_alter_order_clause,
_alter_order_item,
_alter_order_list,
_alter_tablespace_info,
_alter_tablespace_option,
_alter_tablespace_option_list,
_alter_tablespace_options,
_alter_user_list,
_analyze,
_and,
_ascii,
_assign_to_keycache,
_assign_to_keycache_parts,
_attribute,
_begin,
_begin_or_start,
_binlog_base64_event,
_binlog_from,
_binlog_in,
_bit_expr,
_bool_pri,
_btree_or_rtree,
_cache_key_list_or_empty,
_cache_keys_spec,
_call,
_case_stmt_specification,
_cast_type,
_change,
_change_tablespace_access,
_change_tablespace_info,
_change_ts_option,
_change_ts_option_list,
_change_ts_options,
_char,
_charset,
_charset_name,
_charset_name_or_default,
_check,
_check_constraint,
_checksum,
_clear_privileges,
_collation_name,
_collation_name_or_default,
_column_def,
_column_list,
_column_list_id,
_commit,
_comp_op,
_condition_information,
_condition_information_item,
_condition_information_item_name,
_condition_number,
_constraint,
_constraint_key_type,
_create,
_create2,
_create2a,
_create3,
_create_database_option,
_create_database_options,
_create_field_list,
_create_select,
_create_table_option,
_create_table_options,
_create_table_options_space_separated,
_create_view_select,
_create_view_select_paren,
_data_or_xml,
_date_time_type,
_deallocate,
_deallocate_or_drop,
_dec_num,
_dec_num_error,
_default_charset,
_default_collation,
_definer,
_definer_opt,
_definer_tail,
_delete,
_delete_limit_clause,
_delete_option,
_derived_table_list,
_describe,
_describe_command,
_diagnostics_information,
_do,
_drop,
_drop_ts_option,
_drop_ts_options,
_drop_ts_options_list,
_edit,
_else_clause_opt,
_equal,
_esc_table_ref,
_ev_alter_on_schedule_completion,
_ev_ends,
_ev_on_completion,
_ev_schedule_time,
_ev_sql_stmt,
_ev_sql_stmt_inner,
_ev_starts,
_event_tail,
_execute,
_execute_using,
_execute_var_ident,
_execute_var_list,
_explanable_command,
_expr,
_expr_list,
_expr_or_default,
_field_ident,
_field_length,
_field_list,
_field_list_item,
_field_opt_list,
_field_option,
_field_options,
_field_or_var,
_field_spec,
_field_term,
_field_term_list,
_fields,
_fields_or_vars,
_float_options,
_flush,
_flush_option,
_flush_options,
_flush_options_list,
_from_or_in,
_fulltext,
_fulltext_key_opt,
_fulltext_key_options,
_fulltext_key_opts,
_fulltext_options,
_func_datetime_precision,
_function_call_conflict,
_function_call_generic,
_function_call_keyword,
_function_call_nonkeyword,
_geometry_function,
_get_diagnostics,
_get_select_lex,
_gorder_list,
_grant,
_grant_command,
_grant_ident,
_grant_list,
_grant_option,
_grant_option_list,
_grant_options,
_grant_privileges,
_grant_user,
_group_clause,
_group_list,
_handler,
_handler_read_or_scan,
_handler_rkey_function,
_handler_rkey_mode,
_handler_scan_function,
_have_partitioning,
_having_clause,
_help,
_ident,
_ident_eq_list,
_ident_eq_value,
_ident_list,
_ident_list_arg,
_ident_or_empty,
_ident_or_text,
_if_exists,
_ignore_server_id,
_ignore_server_id_list,
_in_sum_expr,
_index_hint_clause,
_index_hint_definition,
_index_hint_type,
_index_hints_list,
_init_key_options,
_insert,
_insert2,
_insert_field_spec,
_insert_ident,
_insert_lock_option,
_insert_table,
_insert_update_elem,
_insert_update_list,
_insert_values,
_install,
_int_type,
_internal_variable_name,
_interval,
_interval_time_stamp,
_into,
_into_destination,
_isolation_level,
_isolation_types,
_join_table,
_join_table_list,
_key_alg,
_key_cache_name,
_key_def,
_key_list,
_key_or_index,
_key_part,
_key_usage_element,
_key_usage_list,
_key_using_alg,
_keycache,
_keycache_list,
_keycache_list_or_parts,
_keys_or_index,
_keyword,
_keyword_sp,
_kill,
_kill_option,
_known_storage_engines,
_label_ident,
_lg_redofile,
_lg_undofile,
_limit_clause,
_limit_option,
_limit_options,
_line_term,
_line_term_list,
_lines_or_rows,
_literal,
_load,
_load_data_lock,
_load_data_set_elem,
_load_data_set_list,
_lock,
_lock_option,
_logfile_group_info,
_logfile_group_name,
_logfile_group_option,
_logfile_group_option_list,
_logfile_group_options,
_master_def,
_master_defs,
_master_file_def,
_master_or_binary,
_merge_insert_types,
_mi_check_type,
_mi_check_types,
_mi_repair_type,
_mi_repair_types,
_nchar,
_no_braces,
_no_definer,
_no_definer_tail,
_normal_join,
_normal_key_opt,
_normal_key_options,
_normal_key_opts,
_normal_key_type,
_not,
_not2,
_now,
_now_or_signed_literal,
_nvarchar,
_object_privilege,
_object_privilege_list,
_olap_opt,
_old_or_new_charset_name,
_old_or_new_charset_name_or_default,
_opt_all,
_opt_and,
_opt_as,
_opt_attribute,
_opt_attribute_list,
_opt_bin_mod,
_opt_binary,
_opt_chain,
_opt_check_constraint,
_opt_checksum_type,
_opt_collate,
_opt_column,
_opt_column_list,
_opt_component,
_opt_constraint,
_opt_create_database_options,
_opt_create_partitioning,
_opt_create_table_options,
_opt_db,
_opt_default,
_opt_delete_option,
_opt_delete_options,
_opt_describe_column,
_opt_distinct,
_opt_duplicate,
_opt_else,
_opt_end_of_input,
_opt_equal,
_opt_escape,
_opt_ev_comment,
_opt_ev_on_completion,
_opt_ev_rename_to,
_opt_ev_sql_stmt,
_opt_ev_status,
_opt_expr,
_opt_expr_list,
_opt_extended_describe,
_opt_field_length,
_opt_field_or_var_spec,
_opt_field_term,
_opt_flush_lock,
_opt_full,
_opt_gconcat_separator,
_opt_gorder_clause,
_opt_grant_option,
_opt_ident,
_opt_if_not_exists,
_opt_ignore,
_opt_ignore_leaves,
_opt_ignore_lines,
_opt_index_hints_list,
_opt_index_lock_algorithm,
_opt_insert_update,
_opt_join_or_resume,
_opt_key_algo,
_opt_key_definition,
_opt_key_or_index,
_opt_key_usage_list,
_opt_limit_clause,
_opt_limit_clause_init,
_opt_line_term,
_opt_linear,
_opt_load_data_charset,
_opt_load_data_set_spec,
_opt_local,
_opt_logfile_group_name,
_opt_low_priority,
_opt_match_clause,
_opt_mi_check_type,
_opt_mi_repair_type,
_opt_migrate,
_opt_natural_language_mode,
_opt_no_write_to_binlog,
_opt_num_parts,
_opt_num_subparts,
_opt_on_update_delete,
_opt_one_phase,
_opt_order_clause,
_opt_outer,
_opt_part_option,
_opt_part_option_list,
_opt_part_options,
_opt_part_values,
_opt_partitioning,
_opt_place,
_opt_precision,
_opt_primary,
_opt_privileges,
_opt_procedure_analyse_params,
_opt_profile_args,
_opt_profile_defs,
_opt_query_expansion,
_opt_query_expression_options,
_opt_ref_list,
_opt_release,
_opt_restrict,
_opt_savepoint,
_opt_select_from,
_opt_set_signal_information,
_opt_signal_value,
_opt_slave_thread_option_list,
_opt_sp_cparam_list,
_opt_sp_cparams,
_opt_start_transaction_option_list,
_opt_storage,
_opt_sub_part,
_opt_sub_partition,
_opt_suspend,
_opt_table,
_opt_table_alias,
_opt_table_list,
_opt_table_options,
_opt_table_sym,
_opt_temporary,
_opt_to,
_opt_ts_autoextend_size,
_opt_ts_comment,
_opt_ts_engine,
_opt_ts_extent_size,
_opt_ts_initial_size,
_opt_ts_max_size,
_opt_ts_nodegroup,
_opt_ts_redo_buffer_size,
_opt_ts_undo_buffer_size,
_opt_udf_expr_list,
_opt_union_order_or_limit,
_opt_unique,
_opt_use_partition,
_opt_value,
_opt_values,
_opt_var_ident_type,
_opt_var_type,
_opt_wild,
_opt_work,
_opt_ws_levels,
_opt_xml_rows_identified_by,
_optimize,
_option_type,
_option_value,
_option_value_following_option_type,
_option_value_list,
_option_value_list_continued,
_option_value_no_option_type,
_optional_braces,
_or,
_order_clause,
_order_dir,
_order_ident,
_order_list,
_order_or_limit,
_param_marker,
_part_column_list,
_part_def_list,
_part_definition,
_part_defs,
_part_field_item,
_part_field_item_list,
_part_field_list,
_part_func,
_part_func_expr,
_part_func_max,
_part_name,
_part_type_def,
_part_value_expr_item,
_part_value_item,
_part_value_item_list,
_part_value_list,
_part_values_in,
_partition,
_partition_entry,
_partitioning,
_precision,
_predicate,
_preload,
_preload_keys,
_preload_keys_parts,
_preload_list,
_preload_list_or_parts,
_prepare,
_prepare_src,
_procedure_analyse_clause,
_procedure_analyse_param,
_profile_def,
_profile_defs,
_purge,
_purge_option,
_purge_options,
_query,
_query_expression_body,
_query_expression_option,
_query_expression_option_list,
_query_specification,
_real_type,
_real_ulong_num,
_real_ulonglong_num,
_ref_list,
_references,
_release,
_remember_end,
_remember_name,
_remove_partitioning,
_rename,
_rename_list,
_reorg_partition_rule,
_reorg_parts_rule,
_repair,
_replace,
_replace_lock_option,
_require_clause,
_require_list,
_require_list_element,
_reset,
_reset_option,
_reset_options,
_resignal_stmt,
_revoke,
_revoke_command,
_rollback,
_row_types,
_savepoint,
_searched_case_stmt,
_searched_when_clause,
_searched_when_clause_list,
_select,
_select_alias,
_select_derived,
_select_derived2,
_select_derived_init,
_select_derived_union,
_select_from,
_select_init,
_select_init2,
_select_init2_derived,
_select_into,
_select_item,
_select_item_list,
_select_lock_type,
_select_option,
_select_option_list,
_select_options,
_select_paren,
_select_paren_derived,
_select_part2,
_select_part2_derived,
_select_var_ident,
_select_var_list,
_select_var_list_init,
_server_def,
_server_option,
_server_options_list,
_set,
_set_expr_or_default,
_sf_tail,
_show,
_show_engine_param,
_show_param,
_signal_allowed_expr,
_signal_condition_information_item_name,
_signal_information_item_list,
_signal_stmt,
_signal_value,
_signed_literal,
_simple_case_stmt,
_simple_expr,
_simple_ident,
_simple_ident_nospvar,
_simple_ident_q,
_simple_target_specification,
_simple_when_clause,
_simple_when_clause_list,
_single_multi,
_size_number,
_slave,
_slave_connection_opts,
_slave_plugin_auth_opt,
_slave_plugin_dir_opt,
_slave_reset_options,
_slave_thread_option,
_slave_thread_option_list,
_slave_until,
_slave_until_opts,
_slave_user_name_opt,
_slave_user_pass_opt,
_sp_a_chistics,
_sp_block_content,
_sp_c_chistic,
_sp_c_chistics,
_sp_chistic,
_sp_cond,
_sp_cparams,
_sp_decl,
_sp_decl_idents,
_sp_decls,
_sp_elseifs,
_sp_fdparam,
_sp_fdparam_list,
_sp_fdparams,
_sp_fetch_list,
_sp_handler_type,
_sp_hcond,
_sp_hcond_element,
_sp_hcond_list,
_sp_if,
_sp_init_param,
_sp_labeled_block,
_sp_labeled_control,
_sp_name,
_sp_opt_default,
_sp_opt_fetch_noise,
_sp_opt_inout,
_sp_opt_label,
_sp_pdparam,
_sp_pdparam_list,
_sp_pdparams,
_sp_proc_stmt,
_sp_proc_stmt_close,
_sp_proc_stmt_fetch,
_sp_proc_stmt_if,
_sp_proc_stmt_iterate,
_sp_proc_stmt_leave,
_sp_proc_stmt_open,
_sp_proc_stmt_return,
_sp_proc_stmt_statement,
_sp_proc_stmt_unlabeled,
_sp_proc_stmts,
_sp_proc_stmts1,
_sp_suid,
_sp_tail,
_sp_unlabeled_block,
_sp_unlabeled_control,
_spatial,
_spatial_key_opt,
_spatial_key_options,
_spatial_key_opts,
_spatial_type,
_sqlstate,
_start,
_start_option_value_list,
_start_option_value_list_following_option_type,
_start_transaction_option,
_start_transaction_option_list,
_statement,
_statement_information,
_statement_information_item,
_statement_information_item_name,
_storage_engines,
_string_list,
_sub_name,
_sub_part_definition,
_sub_part_field_item,
_sub_part_field_list,
_sub_part_func,
_sub_part_list,
_subselect,
_subselect_end,
_subselect_start,
_sum_expr,
_table_alias,
_table_alias_ref,
_table_alias_ref_list,
_table_factor,
_table_ident,
_table_ident_nodb,
_table_ident_opt_wild,
_table_list,
_table_lock,
_table_lock_list,
_table_name,
_table_name_with_opt_use_partition,
_table_option,
_table_options,
_table_or_tables,
_table_ref,
_table_to_table,
_table_to_table_list,
_table_wild,
_table_wild_list,
_table_wild_one,
_tablespace_info,
_tablespace_name,
_tablespace_option,
_tablespace_option_list,
_tablespace_options,
_temporal_literal,
_text_literal,
_text_or_password,
_text_string,
_transaction_access_mode,
_transaction_access_mode_types,
_transaction_characteristics,
_trg_action_time,
_trg_event,
_trigger_tail,
_truncate,
_ts_access_mode,
_ts_datafile,
_ts_wait,
_type,
_type_datetime_precision,
_type_with_opt_collate,
_udf_expr,
_udf_expr_list,
_udf_tail,
_udf_type,
_ulong_num,
_ulonglong_num,
_unicode,
_uninstall,
_union_clause,
_union_list,
_union_opt,
_union_option,
_union_order_or_limit,
_unlock,
_update,
_update_elem,
_update_list,
_use,
_use_partition,
_user,
_user_list,
_using_list,
_values,
_values_list,
_varchar,
_variable,
_variable_aux,
_verb_clause,
_view_algorithm,
_view_check_option,
_view_list,
_view_list_opt,
_view_or_trigger_or_sp_or_event,
_view_replace,
_view_replace_or_algorithm,
_view_select,
_view_select_aux,
_view_suid,
_view_tail,
_when_list,
_where_clause,
_which_area,
_wild_and_where,
_ws_level_flag_desc,
_ws_level_flag_reverse,
_ws_level_flags,
_ws_level_list,
_ws_level_list_item,
_ws_level_list_or_range,
_ws_level_number,
_ws_level_range,
_ws_nweights,
_xa,
_xid,
};

extern const char *symbol_names[];

}

#endif // _SQL_PARSER_SYMBOLS_H_
