#include "sql_parser_symbols.h"
namespace sql
{

const char *symbol_names[]=
{
"",
"123",
"124",
"125",
"126",
"33",
"37",
"38",
"40",
"41",
"42",
"43",
"44",
"45",
"46",
"47",
"58",
"59",
"64",
"94",
"ACCESSIBLE_SYM",
"ACTION",
"ADD",
"ADDDATE_SYM",
"AFTER_SYM",
"AGAINST",
"AGGREGATE_SYM",
"ALGORITHM_SYM",
"ALL",
"ALTER",
"ANALYSE_SYM",
"ANALYZE_SYM",
"AND_AND_SYM",
"AND_SYM",
"ANY_SYM",
"AS",
"ASC",
"ASCII_SYM",
"AT_SYM",
"AUTHORS_SYM",
"AUTOEXTEND_SIZE_SYM",
"AUTO_INC",
"AVG_ROW_LENGTH",
"AVG_SYM",
"BACKUP_SYM",
"BEFORE_SYM",
"BEGIN_SYM",
"BETWEEN_SYM",
"BIGINT",
"BINARY",
"BINLOG_SYM",
"BIN_NUM",
"BIT_AND",
"BIT_OR",
"BIT_SYM",
"BIT_XOR",
"BLOB_SYM",
"BLOCK_SYM",
"BOOLEAN_SYM",
"BOOL_SYM",
"BOTH",
"BTREE_SYM",
"BY",
"BYTE_SYM",
"CACHE_SYM",
"CALL_SYM",
"CASCADE",
"CASCADED",
"CASE_SYM",
"CAST_SYM",
"CATALOG_NAME_SYM",
"CHAIN_SYM",
"CHANGE",
"CHANGED",
"CHARSET",
"CHAR_SYM",
"CHECKSUM_SYM",
"CHECK_SYM",
"CIPHER_SYM",
"CLASS_ORIGIN_SYM",
"CLIENT_SYM",
"CLOSE_SYM",
"COALESCE",
"CODE_SYM",
"COLLATE_SYM",
"COLLATION_SYM",
"COLUMNS",
"COLUMN_FORMAT_SYM",
"COLUMN_NAME_SYM",
"COLUMN_SYM",
"COMMENT_SYM",
"COMMITTED_SYM",
"COMMIT_SYM",
"COMPACT_SYM",
"COMPLETION_SYM",
"COMPRESSED_SYM",
"CONCURRENT",
"CONDITION_SYM",
"CONNECTION_SYM",
"CONSISTENT_SYM",
"CONSTRAINT",
"CONSTRAINT_CATALOG_SYM",
"CONSTRAINT_NAME_SYM",
"CONSTRAINT_SCHEMA_SYM",
"CONTAINS_SYM",
"CONTEXT_SYM",
"CONTINUE_SYM",
"CONTRIBUTORS_SYM",
"CONVERT_SYM",
"COUNT_SYM",
"CPU_SYM",
"CREATE",
"CROSS",
"CUBE_SYM",
"CURDATE",
"CURRENT_SYM",
"CURRENT_USER",
"CURSOR_NAME_SYM",
"CURSOR_SYM",
"CURTIME",
"DATABASE",
"DATABASES",
"DATAFILE_SYM",
"DATA_SYM",
"DATETIME",
"DATE_ADD_INTERVAL",
"DATE_SUB_INTERVAL",
"DATE_SYM",
"DAY_HOUR_SYM",
"DAY_MICROSECOND_SYM",
"DAY_MINUTE_SYM",
"DAY_SECOND_SYM",
"DAY_SYM",
"DEALLOCATE_SYM",
"DECIMAL_NUM",
"DECIMAL_SYM",
"DECLARE_SYM",
"DEFAULT",
"DEFAULT_AUTH_SYM",
"DEFINER_SYM",
"DELAYED_SYM",
"DELAY_KEY_WRITE_SYM",
"DELETE_SYM",
"DESC",
"DESCRIBE",
"DES_KEY_FILE",
"DETERMINISTIC_SYM",
"DIAGNOSTICS_SYM",
"DIRECTORY_SYM",
"DISABLE_SYM",
"DISCARD",
"DISK_SYM",
"DISTINCT",
"DIV_SYM",
"DOUBLE_SYM",
"DO_SYM",
"DROP",
"DUAL_SYM",
"DUMPFILE",
"DUPLICATE_SYM",
"DYNAMIC_SYM",
"EACH_SYM",
"EDIT_SYM",
"ELSE",
"ELSEIF_SYM",
"ENABLE_SYM",
"ENCLOSED",
"END",
"ENDS_SYM",
"END_OF_INPUT",
"ENGINES_SYM",
"ENGINE_SYM",
"ENUM",
"EQ",
"EQUAL_SYM",
"ERRORS",
"ERROR_SYM",
"ESCAPED",
"ESCAPE_SYM",
"EVENTS_SYM",
"EVENT_SYM",
"EVERY_SYM",
"EXCHANGE_SYM",
"EXECUTE_SYM",
"EXISTS",
"EXIT_SYM",
"EXPANSION_SYM",
"EXPIRE_SYM",
"EXPORT_SYM",
"EXTENDED_SYM",
"EXTENT_SIZE_SYM",
"EXTRACT_SYM",
"FALSE_SYM",
"FAST_SYM",
"FAULTS_SYM",
"FETCH_SYM",
"FILE_SYM",
"FIRST_SYM",
"FIXED_SYM",
"FLOAT_NUM",
"FLOAT_SYM",
"FLUSH_SYM",
"FORCE_SYM",
"FOREIGN",
"FORMAT_SYM",
"FOR_SYM",
"FOUND_SYM",
"FROM",
"FULL",
"FULLTEXT_SYM",
"FUNCTION_SYM",
"GE",
"GENERAL",
"GEOMETRYCOLLECTION",
"GEOMETRY_SYM",
"GET_FORMAT",
"GET_SYM",
"GLOBAL_SYM",
"GRANT",
"GRANTS",
"GROUP_CONCAT_SYM",
"GROUP_SYM",
"GT_SYM",
"HANDLER_SYM",
"HASH_SYM",
"HAVING",
"HELP_SYM",
"HEX_NUM",
"HIGH_PRIORITY",
"HOSTS_SYM",
"HOST_SYM",
"HOUR_MICROSECOND_SYM",
"HOUR_MINUTE_SYM",
"HOUR_SECOND_SYM",
"HOUR_SYM",
"IDENT",
"IDENTIFIED_SYM",
"IDENT_QUOTED",
"IDENT_sys",
"IF",
"IGNORE_SERVER_IDS_SYM",
"IGNORE_SYM",
"IMPORT",
"INDEXES",
"INDEX_SYM",
"INFILE",
"INITIAL_SIZE_SYM",
"INNER_SYM",
"INOUT_SYM",
"INSERT",
"INSERT_METHOD",
"INSTALL_SYM",
"INTERVAL_SYM",
"INTO",
"INT_SYM",
"INVOKER_SYM",
"IN_SYM",
"IO_SYM",
"IPC_SYM",
"IS",
"ISOLATION",
"ISSUER_SYM",
"ITERATE_SYM",
"JOIN_SYM",
"KEYS",
"KEY_BLOCK_SIZE",
"KEY_SYM",
"KILL_SYM",
"LANGUAGE_SYM",
"LAST_SYM",
"LE",
"LEADING",
"LEAVES",
"LEAVE_SYM",
"LEFT",
"LESS_SYM",
"LEVEL_SYM",
"LEX_HOSTNAME",
"LIKE",
"LIMIT",
"LINEAR_SYM",
"LINES",
"LINESTRING",
"LIST_SYM",
"LOAD",
"LOCAL_SYM",
"LOCKS_SYM",
"LOCK_SYM",
"LOGFILE_SYM",
"LOGS_SYM",
"LONGBLOB",
"LONGTEXT",
"LONG_NUM",
"LONG_SYM",
"LOOP_SYM",
"LOW_PRIORITY",
"LT",
"MASTER_AUTO_POSITION_SYM",
"MASTER_BIND_SYM",
"MASTER_CONNECT_RETRY_SYM",
"MASTER_DELAY_SYM",
"MASTER_HEARTBEAT_PERIOD_SYM",
"MASTER_HOST_SYM",
"MASTER_LOG_FILE_SYM",
"MASTER_LOG_POS_SYM",
"MASTER_PASSWORD_SYM",
"MASTER_PORT_SYM",
"MASTER_RETRY_COUNT_SYM",
"MASTER_SERVER_ID_SYM",
"MASTER_SSL_CAPATH_SYM",
"MASTER_SSL_CA_SYM",
"MASTER_SSL_CERT_SYM",
"MASTER_SSL_CIPHER_SYM",
"MASTER_SSL_CRLPATH_SYM",
"MASTER_SSL_CRL_SYM",
"MASTER_SSL_KEY_SYM",
"MASTER_SSL_SYM",
"MASTER_SSL_VERIFY_SERVER_CERT_SYM",
"MASTER_SYM",
"MASTER_USER_SYM",
"MATCH",
"MAX_CONNECTIONS_PER_HOUR",
"MAX_QUERIES_PER_HOUR",
"MAX_ROWS",
"MAX_SIZE_SYM",
"MAX_SYM",
"MAX_UPDATES_PER_HOUR",
"MAX_USER_CONNECTIONS_SYM",
"MAX_VALUE_SYM",
"MEDIUMBLOB",
"MEDIUMINT",
"MEDIUMTEXT",
"MEDIUM_SYM",
"MEMORY_SYM",
"MERGE_SYM",
"MESSAGE_TEXT_SYM",
"MICROSECOND_SYM",
"MIGRATE_SYM",
"MINUTE_MICROSECOND_SYM",
"MINUTE_SECOND_SYM",
"MINUTE_SYM",
"MIN_ROWS",
"MIN_SYM",
"MODE_SYM",
"MODIFIES_SYM",
"MODIFY_SYM",
"MOD_SYM",
"MONTH_SYM",
"MULTILINESTRING",
"MULTIPOINT",
"MULTIPOLYGON",
"MUTEX_SYM",
"MYSQL_ERRNO_SYM",
"NAMES_SYM",
"NAME_SYM",
"NATIONAL_SYM",
"NATURAL",
"NCHAR_STRING",
"NCHAR_SYM",
"NDBCLUSTER_SYM",
"NE",
"NEG",
"NEW_SYM",
"NEXT_SYM",
"NODEGROUP_SYM",
"NONE_SYM",
"NOT2_SYM",
"NOT_SYM",
"NOW_SYM",
"NO_SYM",
"NO_WAIT_SYM",
"NO_WRITE_TO_BINLOG",
"NULL_SYM",
"NUM",
"NUMBER_SYM",
"NUMERIC_SYM",
"NUM_literal",
"NVARCHAR_SYM",
"OFFSET_SYM",
"OLD_PASSWORD",
"ON",
"ONE_SYM",
"ONLY_SYM",
"OPEN_SYM",
"OPTIMIZE",
"OPTION",
"OPTIONALLY",
"OPTIONS_SYM",
"OR2_SYM",
"ORDER_SYM",
"OR_OR_SYM",
"OR_SYM",
"OUTER",
"OUTFILE",
"OUT_SYM",
"OWNER_SYM",
"PACK_KEYS_SYM",
"PAGE_SYM",
"PARAM_MARKER",
"PARSER_SYM",
"PARTIAL",
"PARTITIONING_SYM",
"PARTITIONS_SYM",
"PARTITION_SYM",
"PASSWORD",
"PHASE_SYM",
"PLUGINS_SYM",
"PLUGIN_DIR_SYM",
"PLUGIN_SYM",
"POINT_SYM",
"POLYGON",
"PORT_SYM",
"POSITION_SYM",
"PRECISION",
"PREPARE_SYM",
"PRESERVE_SYM",
"PREV_SYM",
"PRIMARY_SYM",
"PRIVILEGES",
"PROCEDURE_SYM",
"PROCESS",
"PROCESSLIST_SYM",
"PROFILES_SYM",
"PROFILE_SYM",
"PROXY_SYM",
"PURGE",
"QUARTER_SYM",
"QUERY_SYM",
"QUICK",
"RANGE_SYM",
"READS_SYM",
"READ_ONLY_SYM",
"READ_SYM",
"READ_WRITE_SYM",
"REAL",
"REBUILD_SYM",
"RECOVER_SYM",
"REDOFILE_SYM",
"REDO_BUFFER_SIZE_SYM",
"REDUNDANT_SYM",
"REFERENCES",
"REGEXP",
"RELAY",
"RELAYLOG_SYM",
"RELAY_LOG_FILE_SYM",
"RELAY_LOG_POS_SYM",
"RELAY_THREAD",
"RELEASE_SYM",
"RELOAD",
"REMOVE_SYM",
"RENAME",
"REORGANIZE_SYM",
"REPAIR",
"REPEATABLE_SYM",
"REPEAT_SYM",
"REPLACE",
"REPLICATION",
"REQUIRE_SYM",
"RESET_SYM",
"RESIGNAL_SYM",
"RESOURCES",
"RESTORE_SYM",
"RESTRICT",
"RESUME_SYM",
"RETURNED_SQLSTATE_SYM",
"RETURNS_SYM",
"RETURN_SYM",
"REVERSE_SYM",
"REVOKE",
"RIGHT",
"ROLLBACK_SYM",
"ROLLUP_SYM",
"ROUTINE_SYM",
"ROWS_SYM",
"ROW_COUNT_SYM",
"ROW_FORMAT_SYM",
"ROW_SYM",
"RTREE_SYM",
"SAVEPOINT_SYM",
"SCHEDULE_SYM",
"SCHEMA_NAME_SYM",
"SECOND_MICROSECOND_SYM",
"SECOND_SYM",
"SECURITY_SYM",
"SELECT_SYM",
"SEPARATOR_SYM",
"SERIALIZABLE_SYM",
"SERIAL_SYM",
"SERVER_SYM",
"SESSION_SYM",
"SET",
"SET_VAR",
"SHARE_SYM",
"SHIFT_LEFT",
"SHIFT_RIGHT",
"SHOW",
"SHUTDOWN",
"SIGNAL_SYM",
"SIGNED_SYM",
"SIMPLE_SYM",
"SLAVE",
"SLOW",
"SMALLINT",
"SNAPSHOT_SYM",
"SOCKET_SYM",
"SONAME_SYM",
"SOUNDS_SYM",
"SOURCE_SYM",
"SPATIAL_SYM",
"SQLEXCEPTION_SYM",
"SQLSTATE_SYM",
"SQLWARNING_SYM",
"SQL_AFTER_GTIDS",
"SQL_AFTER_MTS_GAPS",
"SQL_BEFORE_GTIDS",
"SQL_BIG_RESULT",
"SQL_BUFFER_RESULT",
"SQL_CACHE_SYM",
"SQL_CALC_FOUND_ROWS",
"SQL_NO_CACHE_SYM",
"SQL_SMALL_RESULT",
"SQL_SYM",
"SQL_THREAD",
"SSL_SYM",
"STARTING",
"STARTS_SYM",
"START_SYM",
"STATS_AUTO_RECALC_SYM",
"STATS_PERSISTENT_SYM",
"STATS_SAMPLE_PAGES_SYM",
"STATUS_SYM",
"STDDEV_SAMP_SYM",
"STD_SYM",
"STOP_SYM",
"STORAGE_SYM",
"STRAIGHT_JOIN",
"STRING_SYM",
"SUBCLASS_ORIGIN_SYM",
"SUBDATE_SYM",
"SUBJECT_SYM",
"SUBPARTITIONS_SYM",
"SUBPARTITION_SYM",
"SUBSTRING",
"SUM_SYM",
"SUPER_SYM",
"SUSPEND_SYM",
"SWAPS_SYM",
"SWITCHES_SYM",
"SYSDATE",
"TABLES",
"TABLESPACE",
"TABLE_CHECKSUM_SYM",
"TABLE_NAME_SYM",
"TABLE_REF_PRIORITY",
"TABLE_SYM",
"TEMPORARY",
"TEMPTABLE_SYM",
"TERMINATED",
"TEXT_STRING",
"TEXT_STRING_filesystem",
"TEXT_STRING_literal",
"TEXT_STRING_sys",
"TEXT_STRING_sys_nonewline",
"TEXT_SYM",
"THAN_SYM",
"THEN_SYM",
"TIMESTAMP",
"TIMESTAMP_ADD",
"TIMESTAMP_DIFF",
"TIME_SYM",
"TINYBLOB",
"TINYINT",
"TINYTEXT",
"TO_SYM",
"TRAILING",
"TRANSACTION_SYM",
"TRIGGERS_SYM",
"TRIGGER_SYM",
"TRIM",
"TRUE_SYM",
"TRUNCATE_SYM",
"TYPES_SYM",
"TYPE_SYM",
"UDF_RETURNS_SYM",
"ULONGLONG_NUM",
"UNCOMMITTED_SYM",
"UNDEFINED_SYM",
"UNDERSCORE_CHARSET",
"UNDOFILE_SYM",
"UNDO_BUFFER_SIZE_SYM",
"UNICODE_SYM",
"UNINSTALL_SYM",
"UNION_SYM",
"UNIQUE_SYM",
"UNKNOWN_SYM",
"UNLOCK_SYM",
"UNSIGNED",
"UNTIL_SYM",
"UPDATE_SYM",
"UPGRADE_SYM",
"USAGE",
"USER",
"USE_FRM",
"USE_SYM",
"USING",
"UTC_DATE_SYM",
"UTC_TIMESTAMP_SYM",
"UTC_TIME_SYM",
"VALUES",
"VALUE_SYM",
"VARBINARY",
"VARCHAR",
"VARIABLES",
"VARIANCE_SYM",
"VARYING",
"VAR_SAMP_SYM",
"VIEW_SYM",
"WAIT_SYM",
"WARNINGS",
"WEEK_SYM",
"WEIGHT_STRING_SYM",
"WHEN_SYM",
"WHERE",
"WHILE_SYM",
"WITH",
"WITH_CUBE_SYM",
"WITH_ROLLUP_SYM",
"WORK_SYM",
"WRAPPER_SYM",
"WRITE_SYM",
"X509_SYM",
"XA_SYM",
"XML_SYM",
"XOR",
"YEAR_MONTH_SYM",
"YEAR_SYM",
"ZEROFILL",
"add_column",
"add_log_file",
"add_part_extra",
"add_partition_rule",
"adm_partition",
"all_key_opt",
"all_or_alt_part_name_list",
"all_or_any",
"alt_part_name_item",
"alt_part_name_list",
"alter",
"alter_algorithm_option",
"alter_commands",
"alter_list",
"alter_list_item",
"alter_lock_option",
"alter_logfile_group_info",
"alter_logfile_group_option",
"alter_logfile_group_option_list",
"alter_logfile_group_options",
"alter_order_clause",
"alter_order_item",
"alter_order_list",
"alter_tablespace_info",
"alter_tablespace_option",
"alter_tablespace_option_list",
"alter_tablespace_options",
"alter_user_list",
"analyze",
"and",
"ascii",
"assign_to_keycache",
"assign_to_keycache_parts",
"attribute",
"begin",
"begin_or_start",
"binlog_base64_event",
"binlog_from",
"binlog_in",
"bit_expr",
"bool_pri",
"btree_or_rtree",
"cache_key_list_or_empty",
"cache_keys_spec",
"call",
"case_stmt_specification",
"cast_type",
"change",
"change_tablespace_access",
"change_tablespace_info",
"change_ts_option",
"change_ts_option_list",
"change_ts_options",
"char",
"charset",
"charset_name",
"charset_name_or_default",
"check",
"check_constraint",
"checksum",
"clear_privileges",
"collation_name",
"collation_name_or_default",
"column_def",
"column_list",
"column_list_id",
"commit",
"comp_op",
"condition_information",
"condition_information_item",
"condition_information_item_name",
"condition_number",
"constraint",
"constraint_key_type",
"create",
"create2",
"create2a",
"create3",
"create_database_option",
"create_database_options",
"create_field_list",
"create_select",
"create_table_option",
"create_table_options",
"create_table_options_space_separated",
"create_view_select",
"create_view_select_paren",
"data_or_xml",
"date_time_type",
"deallocate",
"deallocate_or_drop",
"dec_num",
"dec_num_error",
"default_charset",
"default_collation",
"definer",
"definer_opt",
"definer_tail",
"delete",
"delete_limit_clause",
"delete_option",
"derived_table_list",
"describe",
"describe_command",
"diagnostics_information",
"do",
"drop",
"drop_ts_option",
"drop_ts_options",
"drop_ts_options_list",
"edit",
"else_clause_opt",
"equal",
"esc_table_ref",
"ev_alter_on_schedule_completion",
"ev_ends",
"ev_on_completion",
"ev_schedule_time",
"ev_sql_stmt",
"ev_sql_stmt_inner",
"ev_starts",
"event_tail",
"execute",
"execute_using",
"execute_var_ident",
"execute_var_list",
"explanable_command",
"expr",
"expr_list",
"expr_or_default",
"field_ident",
"field_length",
"field_list",
"field_list_item",
"field_opt_list",
"field_option",
"field_options",
"field_or_var",
"field_spec",
"field_term",
"field_term_list",
"fields",
"fields_or_vars",
"float_options",
"flush",
"flush_option",
"flush_options",
"flush_options_list",
"from_or_in",
"fulltext",
"fulltext_key_opt",
"fulltext_key_options",
"fulltext_key_opts",
"fulltext_options",
"func_datetime_precision",
"function_call_conflict",
"function_call_generic",
"function_call_keyword",
"function_call_nonkeyword",
"geometry_function",
"get_diagnostics",
"get_select_lex",
"gorder_list",
"grant",
"grant_command",
"grant_ident",
"grant_list",
"grant_option",
"grant_option_list",
"grant_options",
"grant_privileges",
"grant_user",
"group_clause",
"group_list",
"handler",
"handler_read_or_scan",
"handler_rkey_function",
"handler_rkey_mode",
"handler_scan_function",
"have_partitioning",
"having_clause",
"help",
"ident",
"ident_eq_list",
"ident_eq_value",
"ident_list",
"ident_list_arg",
"ident_or_empty",
"ident_or_text",
"if_exists",
"ignore_server_id",
"ignore_server_id_list",
"in_sum_expr",
"index_hint_clause",
"index_hint_definition",
"index_hint_type",
"index_hints_list",
"init_key_options",
"insert",
"insert2",
"insert_field_spec",
"insert_ident",
"insert_lock_option",
"insert_table",
"insert_update_elem",
"insert_update_list",
"insert_values",
"install",
"int_type",
"internal_variable_name",
"interval",
"interval_time_stamp",
"into",
"into_destination",
"isolation_level",
"isolation_types",
"join_table",
"join_table_list",
"key_alg",
"key_cache_name",
"key_def",
"key_list",
"key_or_index",
"key_part",
"key_usage_element",
"key_usage_list",
"key_using_alg",
"keycache",
"keycache_list",
"keycache_list_or_parts",
"keys_or_index",
"keyword",
"keyword_sp",
"kill",
"kill_option",
"known_storage_engines",
"label_ident",
"lg_redofile",
"lg_undofile",
"limit_clause",
"limit_option",
"limit_options",
"line_term",
"line_term_list",
"lines_or_rows",
"literal",
"load",
"load_data_lock",
"load_data_set_elem",
"load_data_set_list",
"lock",
"lock_option",
"logfile_group_info",
"logfile_group_name",
"logfile_group_option",
"logfile_group_option_list",
"logfile_group_options",
"master_def",
"master_defs",
"master_file_def",
"master_or_binary",
"merge_insert_types",
"mi_check_type",
"mi_check_types",
"mi_repair_type",
"mi_repair_types",
"nchar",
"no_braces",
"no_definer",
"no_definer_tail",
"normal_join",
"normal_key_opt",
"normal_key_options",
"normal_key_opts",
"normal_key_type",
"not",
"not2",
"now",
"now_or_signed_literal",
"nvarchar",
"object_privilege",
"object_privilege_list",
"olap_opt",
"old_or_new_charset_name",
"old_or_new_charset_name_or_default",
"opt_all",
"opt_and",
"opt_as",
"opt_attribute",
"opt_attribute_list",
"opt_bin_mod",
"opt_binary",
"opt_chain",
"opt_check_constraint",
"opt_checksum_type",
"opt_collate",
"opt_column",
"opt_column_list",
"opt_component",
"opt_constraint",
"opt_create_database_options",
"opt_create_partitioning",
"opt_create_table_options",
"opt_db",
"opt_default",
"opt_delete_option",
"opt_delete_options",
"opt_describe_column",
"opt_distinct",
"opt_duplicate",
"opt_else",
"opt_end_of_input",
"opt_equal",
"opt_escape",
"opt_ev_comment",
"opt_ev_on_completion",
"opt_ev_rename_to",
"opt_ev_sql_stmt",
"opt_ev_status",
"opt_expr",
"opt_expr_list",
"opt_extended_describe",
"opt_field_length",
"opt_field_or_var_spec",
"opt_field_term",
"opt_flush_lock",
"opt_full",
"opt_gconcat_separator",
"opt_gorder_clause",
"opt_grant_option",
"opt_ident",
"opt_if_not_exists",
"opt_ignore",
"opt_ignore_leaves",
"opt_ignore_lines",
"opt_index_hints_list",
"opt_index_lock_algorithm",
"opt_insert_update",
"opt_join_or_resume",
"opt_key_algo",
"opt_key_definition",
"opt_key_or_index",
"opt_key_usage_list",
"opt_limit_clause",
"opt_limit_clause_init",
"opt_line_term",
"opt_linear",
"opt_load_data_charset",
"opt_load_data_set_spec",
"opt_local",
"opt_logfile_group_name",
"opt_low_priority",
"opt_match_clause",
"opt_mi_check_type",
"opt_mi_repair_type",
"opt_migrate",
"opt_natural_language_mode",
"opt_no_write_to_binlog",
"opt_num_parts",
"opt_num_subparts",
"opt_on_update_delete",
"opt_one_phase",
"opt_order_clause",
"opt_outer",
"opt_part_option",
"opt_part_option_list",
"opt_part_options",
"opt_part_values",
"opt_partitioning",
"opt_place",
"opt_precision",
"opt_primary",
"opt_privileges",
"opt_procedure_analyse_params",
"opt_profile_args",
"opt_profile_defs",
"opt_query_expansion",
"opt_query_expression_options",
"opt_ref_list",
"opt_release",
"opt_restrict",
"opt_savepoint",
"opt_select_from",
"opt_set_signal_information",
"opt_signal_value",
"opt_slave_thread_option_list",
"opt_sp_cparam_list",
"opt_sp_cparams",
"opt_start_transaction_option_list",
"opt_storage",
"opt_sub_part",
"opt_sub_partition",
"opt_suspend",
"opt_table",
"opt_table_alias",
"opt_table_list",
"opt_table_options",
"opt_table_sym",
"opt_temporary",
"opt_to",
"opt_ts_autoextend_size",
"opt_ts_comment",
"opt_ts_engine",
"opt_ts_extent_size",
"opt_ts_initial_size",
"opt_ts_max_size",
"opt_ts_nodegroup",
"opt_ts_redo_buffer_size",
"opt_ts_undo_buffer_size",
"opt_udf_expr_list",
"opt_union_order_or_limit",
"opt_unique",
"opt_use_partition",
"opt_value",
"opt_values",
"opt_var_ident_type",
"opt_var_type",
"opt_wild",
"opt_work",
"opt_ws_levels",
"opt_xml_rows_identified_by",
"optimize",
"option_type",
"option_value",
"option_value_following_option_type",
"option_value_list",
"option_value_list_continued",
"option_value_no_option_type",
"optional_braces",
"or",
"order_clause",
"order_dir",
"order_ident",
"order_list",
"order_or_limit",
"param_marker",
"part_column_list",
"part_def_list",
"part_definition",
"part_defs",
"part_field_item",
"part_field_item_list",
"part_field_list",
"part_func",
"part_func_expr",
"part_func_max",
"part_name",
"part_type_def",
"part_value_expr_item",
"part_value_item",
"part_value_item_list",
"part_value_list",
"part_values_in",
"partition",
"partition_entry",
"partitioning",
"precision",
"predicate",
"preload",
"preload_keys",
"preload_keys_parts",
"preload_list",
"preload_list_or_parts",
"prepare",
"prepare_src",
"procedure_analyse_clause",
"procedure_analyse_param",
"profile_def",
"profile_defs",
"purge",
"purge_option",
"purge_options",
"query",
"query_expression_body",
"query_expression_option",
"query_expression_option_list",
"query_specification",
"real_type",
"real_ulong_num",
"real_ulonglong_num",
"ref_list",
"references",
"release",
"remember_end",
"remember_name",
"remove_partitioning",
"rename",
"rename_list",
"reorg_partition_rule",
"reorg_parts_rule",
"repair",
"replace",
"replace_lock_option",
"require_clause",
"require_list",
"require_list_element",
"reset",
"reset_option",
"reset_options",
"resignal_stmt",
"revoke",
"revoke_command",
"rollback",
"row_types",
"savepoint",
"searched_case_stmt",
"searched_when_clause",
"searched_when_clause_list",
"select",
"select_alias",
"select_derived",
"select_derived2",
"select_derived_init",
"select_derived_union",
"select_from",
"select_init",
"select_init2",
"select_init2_derived",
"select_into",
"select_item",
"select_item_list",
"select_lock_type",
"select_option",
"select_option_list",
"select_options",
"select_paren",
"select_paren_derived",
"select_part2",
"select_part2_derived",
"select_var_ident",
"select_var_list",
"select_var_list_init",
"server_def",
"server_option",
"server_options_list",
"set",
"set_expr_or_default",
"sf_tail",
"show",
"show_engine_param",
"show_param",
"signal_allowed_expr",
"signal_condition_information_item_name",
"signal_information_item_list",
"signal_stmt",
"signal_value",
"signed_literal",
"simple_case_stmt",
"simple_expr",
"simple_ident",
"simple_ident_nospvar",
"simple_ident_q",
"simple_target_specification",
"simple_when_clause",
"simple_when_clause_list",
"single_multi",
"size_number",
"slave",
"slave_connection_opts",
"slave_plugin_auth_opt",
"slave_plugin_dir_opt",
"slave_reset_options",
"slave_thread_option",
"slave_thread_option_list",
"slave_until",
"slave_until_opts",
"slave_user_name_opt",
"slave_user_pass_opt",
"sp_a_chistics",
"sp_block_content",
"sp_c_chistic",
"sp_c_chistics",
"sp_chistic",
"sp_cond",
"sp_cparams",
"sp_decl",
"sp_decl_idents",
"sp_decls",
"sp_elseifs",
"sp_fdparam",
"sp_fdparam_list",
"sp_fdparams",
"sp_fetch_list",
"sp_handler_type",
"sp_hcond",
"sp_hcond_element",
"sp_hcond_list",
"sp_if",
"sp_init_param",
"sp_labeled_block",
"sp_labeled_control",
"sp_name",
"sp_opt_default",
"sp_opt_fetch_noise",
"sp_opt_inout",
"sp_opt_label",
"sp_pdparam",
"sp_pdparam_list",
"sp_pdparams",
"sp_proc_stmt",
"sp_proc_stmt_close",
"sp_proc_stmt_fetch",
"sp_proc_stmt_if",
"sp_proc_stmt_iterate",
"sp_proc_stmt_leave",
"sp_proc_stmt_open",
"sp_proc_stmt_return",
"sp_proc_stmt_statement",
"sp_proc_stmt_unlabeled",
"sp_proc_stmts",
"sp_proc_stmts1",
"sp_suid",
"sp_tail",
"sp_unlabeled_block",
"sp_unlabeled_control",
"spatial",
"spatial_key_opt",
"spatial_key_options",
"spatial_key_opts",
"spatial_type",
"sqlstate",
"start",
"start_option_value_list",
"start_option_value_list_following_option_type",
"start_transaction_option",
"start_transaction_option_list",
"statement",
"statement_information",
"statement_information_item",
"statement_information_item_name",
"storage_engines",
"string_list",
"sub_name",
"sub_part_definition",
"sub_part_field_item",
"sub_part_field_list",
"sub_part_func",
"sub_part_list",
"subselect",
"subselect_end",
"subselect_start",
"sum_expr",
"table_alias",
"table_alias_ref",
"table_alias_ref_list",
"table_factor",
"table_ident",
"table_ident_nodb",
"table_ident_opt_wild",
"table_list",
"table_lock",
"table_lock_list",
"table_name",
"table_name_with_opt_use_partition",
"table_option",
"table_options",
"table_or_tables",
"table_ref",
"table_to_table",
"table_to_table_list",
"table_wild",
"table_wild_list",
"table_wild_one",
"tablespace_info",
"tablespace_name",
"tablespace_option",
"tablespace_option_list",
"tablespace_options",
"temporal_literal",
"text_literal",
"text_or_password",
"text_string",
"transaction_access_mode",
"transaction_access_mode_types",
"transaction_characteristics",
"trg_action_time",
"trg_event",
"trigger_tail",
"truncate",
"ts_access_mode",
"ts_datafile",
"ts_wait",
"type",
"type_datetime_precision",
"type_with_opt_collate",
"udf_expr",
"udf_expr_list",
"udf_tail",
"udf_type",
"ulong_num",
"ulonglong_num",
"unicode",
"uninstall",
"union_clause",
"union_list",
"union_opt",
"union_option",
"union_order_or_limit",
"unlock",
"update",
"update_elem",
"update_list",
"use",
"use_partition",
"user",
"user_list",
"using_list",
"values",
"values_list",
"varchar",
"variable",
"variable_aux",
"verb_clause",
"view_algorithm",
"view_check_option",
"view_list",
"view_list_opt",
"view_or_trigger_or_sp_or_event",
"view_replace",
"view_replace_or_algorithm",
"view_select",
"view_select_aux",
"view_suid",
"view_tail",
"when_list",
"where_clause",
"which_area",
"wild_and_where",
"ws_level_flag_desc",
"ws_level_flag_reverse",
"ws_level_flags",
"ws_level_list",
"ws_level_list_item",
"ws_level_list_or_range",
"ws_level_number",
"ws_level_range",
"ws_nweights",
"xa",
"xid",
};

} // namespace sql_parser_symbols
