//
// This header file auto-generated for the template
//    c:\Users\vkolesnikov\co\mysql-workbench-pro\modules\db.mysql\res\reporting_includes\.\..\reporting\Basic_Text.tpl\basic_text_report.txt.tpl
// by ..\..\..\..\..\mysql-gui-win-res\bin\make_tpl_varname_h
// DO NOT MODIFY THIS FILE DIRECTLY
//
const char* const kbtr_CREATE_SCHEMA = "CREATE_SCHEMA";
const char* const kbtr_CREATE_SCHEMA_NAME = "CREATE_SCHEMA_NAME";
const char* const kbtr_DROP_SCHEMA = "DROP_SCHEMA";
const char* const kbtr_DROP_SCHEMA_NAME = "DROP_SCHEMA_NAME";
const char* const kbtr_ALTER_SCHEMA = "ALTER_SCHEMA";
const char* const kbtr_ALTER_SCHEMA_NAME = "ALTER_SCHEMA_NAME";
const char* const kbtr_OLD_SCHEMA_NAME = "OLD_SCHEMA_NAME";
const char* const kbtr_NEW_SCHEMA_NAME = "NEW_SCHEMA_NAME";
const char* const kbtr_ALTER_SCHEMA_CHARSET = "ALTER_SCHEMA_CHARSET";
const char* const kbtr_OLD_SCHEMA_CHARSET = "OLD_SCHEMA_CHARSET";
const char* const kbtr_NEW_SCHEMA_CHARSET = "NEW_SCHEMA_CHARSET";
const char* const kbtr_ALTER_SCHEMA_COLLATE = "ALTER_SCHEMA_COLLATE";
const char* const kbtr_OLD_SCHEMA_COLLATE = "OLD_SCHEMA_COLLATE";
const char* const kbtr_NEW_SCHEMA_COLLATE = "NEW_SCHEMA_COLLATE";
const char* const kbtr_DROP_TABLE = "DROP_TABLE";
const char* const kbtr_DROP_TABLE_NAME = "DROP_TABLE_NAME";
const char* const kbtr_CREATE_TABLE = "CREATE_TABLE";
const char* const kbtr_CREATE_TABLE_NAME = "CREATE_TABLE_NAME";
const char* const kbtr_CREATE_TABLE_COLUMNS_HEADER = "CREATE_TABLE_COLUMNS_HEADER";
const char* const kbtr_TABLE_COLUMN = "TABLE_COLUMN";
const char* const kbtr_TABLE_COLUMN_NAME = "TABLE_COLUMN_NAME";
const char* const kbtr_TABLE_COLUMN_TYPE = "TABLE_COLUMN_TYPE";
const char* const kbtr_CREATE_TABLE_COLUMNS_FOOTER = "CREATE_TABLE_COLUMNS_FOOTER";
const char* const kbtr_CREATE_TABLE_INDEXES_HEADER = "CREATE_TABLE_INDEXES_HEADER";
const char* const kbtr_TABLE_INDEX = "TABLE_INDEX";
const char* const kbtr_TABLE_INDEX_NAME = "TABLE_INDEX_NAME";
const char* const kbtr_TABLE_INDEX_COLUMNS = "TABLE_INDEX_COLUMNS";
const char* const kbtr_CREATE_TABLE_INDEXES_FOOTER = "CREATE_TABLE_INDEXES_FOOTER";
const char* const kbtr_CREATE_TABLE_FKS_HEADER = "CREATE_TABLE_FKS_HEADER";
const char* const kbtr_TABLE_FK = "TABLE_FK";
const char* const kbtr_TABLE_FK_NAME = "TABLE_FK_NAME";
const char* const kbtr_TABLE_FK_COLUMNS = "TABLE_FK_COLUMNS";
const char* const kbtr_TABLE_FK_REF_TABLE = "TABLE_FK_REF_TABLE";
const char* const kbtr_TABLE_FK_REF_COLUMNS = "TABLE_FK_REF_COLUMNS";
const char* const kbtr_TABLE_FK_ON_UPDATE = "TABLE_FK_ON_UPDATE";
const char* const kbtr_TABLE_FK_ON_DELETE = "TABLE_FK_ON_DELETE";
const char* const kbtr_CREATE_TABLE_FKS_FOOTER = "CREATE_TABLE_FKS_FOOTER";
const char* const kbtr_CREATE_TABLE_ATTRIBUTES_HEADER = "CREATE_TABLE_ATTRIBUTES_HEADER";
const char* const kbtr_TABLE_ATTR_ENGINE = "TABLE_ATTR_ENGINE";
const char* const kbtr_TABLE_ENGINE = "TABLE_ENGINE";
const char* const kbtr_TABLE_ATTR_AUTOINC = "TABLE_ATTR_AUTOINC";
const char* const kbtr_TABLE_AUTOINC = "TABLE_AUTOINC";
const char* const kbtr_TABLE_ATTR_PASSWORD = "TABLE_ATTR_PASSWORD";
const char* const kbtr_TABLE_PASSWORD = "TABLE_PASSWORD";
const char* const kbtr_TABLE_ATTR_DELAY_KEY_WRITE = "TABLE_ATTR_DELAY_KEY_WRITE";
const char* const kbtr_TABLE_DELAY_KEY_WRITE = "TABLE_DELAY_KEY_WRITE";
const char* const kbtr_TABLE_ATTR_CHARSET = "TABLE_ATTR_CHARSET";
const char* const kbtr_TABLE_CHARSET = "TABLE_CHARSET";
const char* const kbtr_TABLE_ATTR_COLLATE = "TABLE_ATTR_COLLATE";
const char* const kbtr_TABLE_COLLATE = "TABLE_COLLATE";
const char* const kbtr_TABLE_ATTR_MERGE_UNION = "TABLE_ATTR_MERGE_UNION";
const char* const kbtr_TABLE_MERGE_UNION = "TABLE_MERGE_UNION";
const char* const kbtr_TABLE_ATTR_MERGE_INSERT = "TABLE_ATTR_MERGE_INSERT";
const char* const kbtr_TABLE_MERGE_INSERT = "TABLE_MERGE_INSERT";
const char* const kbtr_TABLE_ATTR_PACK_KEYS = "TABLE_ATTR_PACK_KEYS";
const char* const kbtr_TABLE_PACK_KEYS = "TABLE_PACK_KEYS";
const char* const kbtr_TABLE_ATTR_CHECKSUM = "TABLE_ATTR_CHECKSUM";
const char* const kbtr_TABLE_CHECKSUM = "TABLE_CHECKSUM";
const char* const kbtr_TABLE_ATTR_ROW_FORMAT = "TABLE_ATTR_ROW_FORMAT";
const char* const kbtr_TABLE_ROW_FORMAT = "TABLE_ROW_FORMAT";
const char* const kbtr_TABLE_ATTR_KEY_BLOCK_SIZE = "TABLE_ATTR_KEY_BLOCK_SIZE";
const char* const kbtr_TABLE_KEY_BLOCK_SIZE = "TABLE_KEY_BLOCK_SIZE";
const char* const kbtr_TABLE_ATTR_AVG_ROW_LENGTH = "TABLE_ATTR_AVG_ROW_LENGTH";
const char* const kbtr_TABLE_AVG_ROW_LENGTH = "TABLE_AVG_ROW_LENGTH";
const char* const kbtr_TABLE_ATTR_MIN_ROWS = "TABLE_ATTR_MIN_ROWS";
const char* const kbtr_TABLE_MIN_ROWS = "TABLE_MIN_ROWS";
const char* const kbtr_TABLE_ATTR_MAX_ROWS = "TABLE_ATTR_MAX_ROWS";
const char* const kbtr_TABLE_MAX_ROWS = "TABLE_MAX_ROWS";
const char* const kbtr_TABLE_ATTR_COMMENT = "TABLE_ATTR_COMMENT";
const char* const kbtr_TABLE_COMMENT = "TABLE_COMMENT";
const char* const kbtr_TABLE_ATTR_DATADIR = "TABLE_ATTR_DATADIR";
const char* const kbtr_TABLE_DATADIR = "TABLE_DATADIR";
const char* const kbtr_TABLE_ATTR_INDEXDIR = "TABLE_ATTR_INDEXDIR";
const char* const kbtr_TABLE_INDEXDIR = "TABLE_INDEXDIR";
const char* const kbtr_CREATE_TABLE_ATTRIBUTES_FOOTER = "CREATE_TABLE_ATTRIBUTES_FOOTER";
const char* const kbtr_ALTER_TABLE = "ALTER_TABLE";
const char* const kbtr_ALTER_TABLE_NAME = "ALTER_TABLE_NAME";
const char* const kbtr_ALTER_TABLE_COLUMNS_HEADER = "ALTER_TABLE_COLUMNS_HEADER";
const char* const kbtr_TABLE_COLUMN_ADDED = "TABLE_COLUMN_ADDED";
const char* const kbtr_TABLE_COLUMN_REMOVED = "TABLE_COLUMN_REMOVED";
const char* const kbtr_TABLE_COLUMN_MODIFIED = "TABLE_COLUMN_MODIFIED";
const char* const kbtr_ALTER_TABLE_COLUMNS_FOOTER = "ALTER_TABLE_COLUMNS_FOOTER";
const char* const kbtr_ALTER_TABLE_INDEXES_HEADER = "ALTER_TABLE_INDEXES_HEADER";
const char* const kbtr_TABLE_INDEX_ADDED = "TABLE_INDEX_ADDED";
const char* const kbtr_TABLE_INDEX_REMOVED = "TABLE_INDEX_REMOVED";
const char* const kbtr_TABLE_INDEX_MODIFIED = "TABLE_INDEX_MODIFIED";
const char* const kbtr_ALTER_TABLE_INDEXES_FOOTER = "ALTER_TABLE_INDEXES_FOOTER";
const char* const kbtr_ALTER_TABLE_FKS_HEADER = "ALTER_TABLE_FKS_HEADER";
const char* const kbtr_TABLE_FK_ADDED = "TABLE_FK_ADDED";
const char* const kbtr_TABLE_FK_REMOVED = "TABLE_FK_REMOVED";
const char* const kbtr_ALTER_TABLE_FKS_FOOTER = "ALTER_TABLE_FKS_FOOTER";
const char* const kbtr_ALTER_TABLE_ATTRIBUTES_HEADER = "ALTER_TABLE_ATTRIBUTES_HEADER";
const char* const kbtr_TABLE_ATTR_NAME = "TABLE_ATTR_NAME";
const char* const kbtr_OLD_TABLE_NAME = "OLD_TABLE_NAME";
const char* const kbtr_NEW_TABLE_NAME = "NEW_TABLE_NAME";
const char* const kbtr_OLD_TABLE_ENGINE = "OLD_TABLE_ENGINE";
const char* const kbtr_NEW_TABLE_ENGINE = "NEW_TABLE_ENGINE";
const char* const kbtr_OLD_TABLE_AUTOINC = "OLD_TABLE_AUTOINC";
const char* const kbtr_NEW_TABLE_AUTOINC = "NEW_TABLE_AUTOINC";
const char* const kbtr_OLD_TABLE_PASSWORD = "OLD_TABLE_PASSWORD";
const char* const kbtr_NEW_TABLE_PASSWORD = "NEW_TABLE_PASSWORD";
const char* const kbtr_OLD_TABLE_DELAY_KEY_WRITE = "OLD_TABLE_DELAY_KEY_WRITE";
const char* const kbtr_NEW_TABLE_DELAY_KEY_WRITE = "NEW_TABLE_DELAY_KEY_WRITE";
const char* const kbtr_OLD_TABLE_CHARSET = "OLD_TABLE_CHARSET";
const char* const kbtr_NEW_TABLE_CHARSET = "NEW_TABLE_CHARSET";
const char* const kbtr_OLD_TABLE_COLLATE = "OLD_TABLE_COLLATE";
const char* const kbtr_NEW_TABLE_COLLATE = "NEW_TABLE_COLLATE";
const char* const kbtr_OLD_TABLE_MERGE_UNION = "OLD_TABLE_MERGE_UNION";
const char* const kbtr_NEW_TABLE_MERGE_UNION = "NEW_TABLE_MERGE_UNION";
const char* const kbtr_OLD_TABLE_MERGE_INSERT = "OLD_TABLE_MERGE_INSERT";
const char* const kbtr_NEW_TABLE_MERGE_INSERT = "NEW_TABLE_MERGE_INSERT";
const char* const kbtr_OLD_TABLE_PACK_KEYS = "OLD_TABLE_PACK_KEYS";
const char* const kbtr_NEW_TABLE_PACK_KEYS = "NEW_TABLE_PACK_KEYS";
const char* const kbtr_OLD_TABLE_CHECKSUM = "OLD_TABLE_CHECKSUM";
const char* const kbtr_NEW_TABLE_CHECKSUM = "NEW_TABLE_CHECKSUM";
const char* const kbtr_OLD_TABLE_ROW_FORMAT = "OLD_TABLE_ROW_FORMAT";
const char* const kbtr_NEW_TABLE_ROW_FORMAT = "NEW_TABLE_ROW_FORMAT";
const char* const kbtr_OLD_TABLE_KEY_BLOCK_SIZE = "OLD_TABLE_KEY_BLOCK_SIZE";
const char* const kbtr_NEW_TABLE_KEY_BLOCK_SIZE = "NEW_TABLE_KEY_BLOCK_SIZE";
const char* const kbtr_OLD_TABLE_AVG_ROW_LENGTH = "OLD_TABLE_AVG_ROW_LENGTH";
const char* const kbtr_NEW_TABLE_AVG_ROW_LENGTH = "NEW_TABLE_AVG_ROW_LENGTH";
const char* const kbtr_OLD_TABLE_MIN_ROWS = "OLD_TABLE_MIN_ROWS";
const char* const kbtr_NEW_TABLE_MIN_ROWS = "NEW_TABLE_MIN_ROWS";
const char* const kbtr_OLD_TABLE_MAX_ROWS = "OLD_TABLE_MAX_ROWS";
const char* const kbtr_NEW_TABLE_MAX_ROWS = "NEW_TABLE_MAX_ROWS";
const char* const kbtr_OLD_TABLE_COMMENT = "OLD_TABLE_COMMENT";
const char* const kbtr_NEW_TABLE_COMMENT = "NEW_TABLE_COMMENT";
const char* const kbtr_ALTER_TABLE_ATTRIBUTES_FOOTER = "ALTER_TABLE_ATTRIBUTES_FOOTER";
const char* const kbtr_ALTER_TABLE_PART_HEADER = "ALTER_TABLE_PART_HEADER";
const char* const kbtr_ALTER_TABLE_PART_ADDED = "ALTER_TABLE_PART_ADDED";
const char* const kbtr_ALTER_TABLE_PART_MODIFIED = "ALTER_TABLE_PART_MODIFIED";
const char* const kbtr_ALTER_TABLE_PART_REMOVED = "ALTER_TABLE_PART_REMOVED";
const char* const kbtr_ALTER_TABLE_PART_FOOTER = "ALTER_TABLE_PART_FOOTER";
const char* const kbtr_DROP_TRIGGER = "DROP_TRIGGER";
const char* const kbtr_DROP_TRIGGER_NAME = "DROP_TRIGGER_NAME";
const char* const kbtr_CREATE_TRIGGER = "CREATE_TRIGGER";
const char* const kbtr_CREATE_TRIGGER_NAME = "CREATE_TRIGGER_NAME";
const char* const kbtr_DROP_VIEW = "DROP_VIEW";
const char* const kbtr_DROP_VIEW_NAME = "DROP_VIEW_NAME";
const char* const kbtr_CREATE_VIEW = "CREATE_VIEW";
const char* const kbtr_CREATE_VIEW_NAME = "CREATE_VIEW_NAME";
const char* const kbtr_DROP_ROUTINE = "DROP_ROUTINE";
const char* const kbtr_DROP_ROUTINE_NAME = "DROP_ROUTINE_NAME";
const char* const kbtr_CREATE_ROUTINE = "CREATE_ROUTINE";
const char* const kbtr_CREATE_ROUTINE_NAME = "CREATE_ROUTINE_NAME";
const char* const kbtr_DROP_USER = "DROP_USER";
const char* const kbtr_DROP_USER_NAME = "DROP_USER_NAME";
const char* const kbtr_CREATE_USER = "CREATE_USER";
const char* const kbtr_CREATE_USER_NAME = "CREATE_USER_NAME";
