/*
 * Copyright (c) 2014, 2018, Oracle and/or its affiliates. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License, version 2.0,
 * as published by the Free Software Foundation.
 *
 * This program is designed to work with certain software (including
 * but not limited to OpenSSL) that is licensed under separate terms, as
 * designated in a particular file or component or in included license
 * documentation.  The authors of MySQL hereby grant you an additional
 * permission to link the program and your derivative works with the
 * separately licensed software that they have either included with
 * the program or referenced in the documentation.
 * This program is distributed in the hope that it will be useful,  but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License, version 2.0, for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA 
 */

#ifndef _GEOM_DRAW_BOX_H_
#define _GEOM_DRAW_BOX_H_

#include <mforms/drawbox.h>
#include <ogr_geometry.h>
#include "wbpublic_public_interface.h"

class WBPUBLICBACKEND_PUBLIC_FUNC GeomDrawBox : public mforms::DrawBox {
  OGRGeometry *_geom;
  int _srid;

public:
  GeomDrawBox() : _geom(NULL), _srid(0) {
  }

  void draw_ring(cairo_t *cr, OGRRawPoint *points, int num_points, double scale, double x, double y, double height);

  void draw_ring_vertices(cairo_t *cr, OGRRawPoint *points, int num_points, double scale, double x, double y,
                          double height);

  void draw_geometry(cairo_t *cr, OGRGeometry *geom, double scale, double x, double y, double height);

  void draw_polygon(cairo_t *cr, OGRPolygon *poly, double scale, double x, double y, double height);
  void set_data(const std::string &text);

  virtual void repaint(cairo_t *cr, int x, int y, int w, int h);

  int getSrid() const;
};

#endif
