/*
 * Copyright (c) 2007, 2018, Oracle and/or its affiliates. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License, version 2.0,
 * as published by the Free Software Foundation.
 *
 * This program is designed to work with certain software (including
 * but not limited to OpenSSL) that is licensed under separate terms, as
 * designated in a particular file or component or in included license
 * documentation.  The authors of MySQL hereby grant you an additional
 * permission to link the program and your derivative works with the
 * separately licensed software that they have either included with
 * the program or referenced in the documentation.
 * This program is distributed in the hope that it will be useful,  but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License, version 2.0, for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA 
 */

#pragma once

#include "grt.h"
#include "grt/tree_model.h"

namespace bec {

  class CharsetList : public ListModel {
  public:
    CharsetList(const std::string &path);

    enum CharsetListColumns { Name };

    virtual size_t count_children(const NodeId &parent);
    virtual bool get_field(const NodeId &node, ColumnId column, std::string &value);
    virtual std::string get_field_description(const NodeId &node, ColumnId column);
    void picked_charset(const NodeId &node);

  protected:
    std::list<size_t> _recently_used;
    std::string _charset_list_path;
  };
};
