/*
 * Copyright (c) 2009, 2018, Oracle and/or its affiliates. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License, version 2.0,
 * as published by the Free Software Foundation.
 *
 * This program is designed to work with certain software (including
 * but not limited to OpenSSL) that is licensed under separate terms, as
 * designated in a particular file or component or in included license
 * documentation.  The authors of MySQL hereby grant you an additional
 * permission to link the program and your derivative works with the
 * separately licensed software that they have either included with
 * the program or referenced in the documentation.
 * This program is distributed in the hope that it will be useful,  but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License, version 2.0, for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA 
 */

#ifndef _DB_REV_ENG_BE_H_
#define _DB_REV_ENG_BE_H_

#include "sql_import_be.h"

#include "db_mysql_public_interface.h"
#include "db_plugin_be.h"

class WBPLUGINDBMYSQLBE_PUBLIC_FUNC Db_rev_eng : public Db_plugin, public Sql_import {
private:
  std::string task_desc();
  void parse_sql_script(parsers::MySQLParserServices::Ref sql_parser, parsers::MySQLParserContext::Ref context,
                        db_CatalogRef &catalog, const std::string &sql_scrtipt, grt::DictRef &options);
  db_CatalogRef target_catalog();

public:
  std::string sql_script();
  void sql_script(const std::string &sql_script) {
    Db_plugin::sql_script(sql_script);
  }

public:
  Db_rev_eng() : Db_plugin(), Sql_import() {
  }
  void grtm();
  GrtVersionRef getVersion();
};

#endif /* _DB_REV_ENG_BE_H_ */
