/*
 * Copyright (c) 2009, 2018, Oracle and/or its affiliates. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License, version 2.0,
 * as published by the Free Software Foundation.
 *
 * This program is designed to work with certain software (including
 * but not limited to OpenSSL) that is licensed under separate terms, as
 * designated in a particular file or component or in included license
 * documentation.  The authors of MySQL hereby grant you an additional
 * permission to link the program and your derivative works with the
 * separately licensed software that they have either included with
 * the program or referenced in the documentation.
 * This program is distributed in the hope that it will be useful,  but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License, version 2.0, for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA 
 */

#pragma once

#include "mforms/form.h"
#include "grt.h"
#include "wbpublic_public_interface.h"

namespace mforms {
  class Selector;
  class Button;
}

class WBPUBLICBACKEND_PUBLIC_FUNC FileCharsetDialog : public mforms::Form {
  mforms::Selector *_charset;
  mforms::Button *_ok;
  mforms::Button *_cancel;
  mforms::Button *_run;

  bool _run_clicked;

  FileCharsetDialog(const std::string &title, const std::string &message);

  void run_clicked();

public:
  enum Result { Cancelled, Accepted, RunInstead };

  std::string run(const std::string &default_encoding);

  static Result ensure_filedata_utf8(const char *data, size_t length, const std::string &encoding,
                                     const std::string &filename, char *&utf8_data,
                                     std::string *original_encoding = nullptr);
};
